
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

// distant.cpp*
#include "pbrt.h"
#include "light.h"
#include "shape.h"
#include "scene.h"
#include "sunsky.dat"

#define radians(x) ((x)/180.0*M_PI)

const struct spectrum41_t
{
	int start_wavelength;
	int end_wavelength;
	int nsamples;

	float samples[41];
	float x, y, z;
	float r, g, b;
};

// DistantLight Declarations
class DistantLight : public Light {
public:
	// DistantLight Public Methods
	DistantLight(const Transform &light2world, const Spectrum &radiance, const Vector &dir, const float &turb, const float &thetaS);
	bool IsDeltaLight() const { return true; }
	Spectrum Sample_L(const Point &p, Vector *wi, VisibilityTester *) const;
	Spectrum Power(const Scene *scene) const {
		Point worldCenter;
		float worldRadius;
		scene->WorldBound().BoundingSphere(&worldCenter,
		                                   &worldRadius);
		return L * M_PI * worldRadius * worldRadius;
	}
	Spectrum Sample_L(const Point &P, float u1, float u2, Vector *wo, float *pdf,
		VisibilityTester *visibility) const;
	Spectrum Sample_L(const Scene *scene, float u1, float u2,
		float u3, float u4, Ray *ray, float *pdf) const;
	float Pdf(const Point &, const Vector &) const;
	
	// Directional sunlight methods
	void DistantLight::ComputeSunlight() const;
	void DistantLight::spectrum_to_xyz(double *x, double *y, double *z) const;
	void DistantLight::xyz_to_rgb(double xc, double yc, double zc,
                double *r, double *g, double *b) const;
	void HSVtoRGB( float *r, float *g, float *b, float h, float s, float v ) const;
	void RGBtoHSV( float r, float g, float b, float *h, float *s, float *v ) const;
private:
	// DistantLight Private Data
	Vector lightDir;
	Spectrum L;
	float turbidity;
	float sun_theta;
	mutable spectrum41_t spectrum41;
};
// DistantLight Method Definitions
DistantLight::DistantLight(const Transform &light2world,
		const Spectrum &radiance, const Vector &dir, const float &turb, const float &thetaS)
	: Light(light2world) {
	lightDir = Normalize(LightToWorld(dir));
	L = radiance;
	turbidity = turb;
	sun_theta = thetaS;

	ComputeSunlight();
	float rgb[3];
	rgb[0] = spectrum41.r;
	rgb[1] = spectrum41.g;
	rgb[2] = spectrum41.b;
	L = Spectrum(rgb);
}
Spectrum DistantLight::Sample_L(const Point &p,
		Vector *wi, VisibilityTester *visibility) const {
	*wi = lightDir;
	visibility->SetRay(p, *wi);
	return L;
}
Spectrum DistantLight::Sample_L(const Point &p, float u1, float u2,
		Vector *wi, float *pdf, VisibilityTester *visibility) const {
	*pdf = 1.f;
	return Sample_L(p, wi, visibility);
}
float DistantLight::Pdf(const Point &, const Vector &) const {
	return 0.;
}
Spectrum DistantLight::Sample_L(const Scene *scene,
		float u1, float u2, float u3, float u4,
		Ray *ray, float *pdf) const {
	// Choose point on disk oriented toward infinite light direction
	Point worldCenter;
	float worldRadius;
	scene->WorldBound().BoundingSphere(&worldCenter,
	                                   &worldRadius);
	Vector v1, v2;
	CoordinateSystem(lightDir, &v1, &v2);
	float d1, d2;
	ConcentricSampleDisk(u1, u2, &d1, &d2);
	Point Pdisk =
		worldCenter + worldRadius * (d1 * v1 + d2 * v2);
	// Set ray origin and direction for infinite light ray
	ray->o = Pdisk + worldRadius * lightDir;
	ray->d = -lightDir;
	*pdf = 1.f / (M_PI * worldRadius * worldRadius);
	return L;
}

void DistantLight::ComputeSunlight() const  {
	const double alpha  = 1.3;
	const double lozone = 0.35; 
	const double w = 2.0;

	double beta;
	double tauR, tauA, tauO, tauG, tauWA;
	double m;
	double lambda;
	double amp;

	beta = 0.04608365822050 * turbidity - 0.04586025928522;

	// Relative optical mass
	m = 1.0 / (cos(sun_theta) + 0.15 * pow(93.885 - sun_theta / M_PI * 180.0, -1.253));

	if (sun_theta < 0.5 * M_PI) {
		for (int i=0, lambda=380; lambda<781; i++, lambda += 10) {

			// Rayleigh scattering
			tauR = exp(-m * 0.008735 * pow(lambda/1000.0, -4.08));
			// Aerosol (water + dust) attenuation
			tauA = exp(-m * beta * pow(lambda/1000.0, -alpha));
			// Attenuation due to ozone absorption
			tauO = exp(-m * k_oAmplitudes[i] * lozone);
			// Attenuation due to mixed gases absorption
			tauG = exp(-1.41 * k_gAmplitudes[i] * m / pow(1.0 + 118.93 * k_gAmplitudes[i] * m, 0.45));
			// Attenuation due to water vapor absorption
			tauWA = exp(-0.2385 * k_waAmplitudes[i] * w * m /
					pow(1.0 + 20.07 * k_waAmplitudes[i] * w * m, 0.45));
			// 100.0 comes from solAmplitudes begin in wrong units.
			amp = 100.0 * solAmplitudes[i] * tauR * tauA * tauO * tauG * tauWA;
			spectrum41.samples[i] = amp;
		}
	}
	else { // Under the horizon
		for (int i = 0; i < spectrum41.nsamples; i++) {
			spectrum41.samples[i] = 0.0;
		}
	}

	// from Spectrum to XYZ
	double X, Y, Z;
	spectrum_to_xyz(&X, &Y, &Z);

	// from XYZ to RGB
	float rgb[3];
	rgb[0] =  3.240479f * X + -1.537150f * Y + -0.498535f * Z;
	rgb[1] =  -0.969256f * X +  1.875991f * Y +  0.041556f * Z;
	rgb[2] =  0.055648f * X + -0.204043f * Y +  1.057311f * Z;

	// RGB -> HSV -> RGB to get the best luminance scaling (in HSV)
	float expFactor = 18.0f;
	float hsv[3];
	RGBtoHSV(rgb[0], rgb[1], rgb[2], &hsv[0], &hsv[1], &hsv[2]);
	hsv[2] = 1.0f - exp(-expFactor * hsv[2]);
	HSVtoRGB(&rgb[0], &rgb[1], &rgb[2], hsv[0], hsv[1], hsv[2]);	
	
	spectrum41.r = rgb[0];
	spectrum41.g = rgb[1];
	spectrum41.b = rgb[2];

	/*xyz_to_rgb(sun_x, sun_y, sun_z, &sun_r, &sun_g, &sun_b);
	spectrum41.x = sun_x;
	spectrum41.y = sun_y;
	spectrum41.z = sun_z;
	spectrum41.r = sun_r;
	spectrum41.g = sun_g;
	spectrum41.b = sun_b;*/
}

void DistantLight::spectrum_to_xyz(double *x, double *y, double *z) const {
    int i;
    double lambda, X = 0, Y = 0, Z = 0, XYZ;

    static double cie_colour_match[81][3] = {
        {0.0014,0.0000,0.0065}, {0.0022,0.0001,0.0105}, {0.0042,0.0001,0.0201},
        {0.0076,0.0002,0.0362}, {0.0143,0.0004,0.0679}, {0.0232,0.0006,0.1102},
        {0.0435,0.0012,0.2074}, {0.0776,0.0022,0.3713}, {0.1344,0.0040,0.6456},
        {0.2148,0.0073,1.0391}, {0.2839,0.0116,1.3856}, {0.3285,0.0168,1.6230},
        {0.3483,0.0230,1.7471}, {0.3481,0.0298,1.7826}, {0.3362,0.0380,1.7721},
        {0.3187,0.0480,1.7441}, {0.2908,0.0600,1.6692}, {0.2511,0.0739,1.5281},
        {0.1954,0.0910,1.2876}, {0.1421,0.1126,1.0419}, {0.0956,0.1390,0.8130},
        {0.0580,0.1693,0.6162}, {0.0320,0.2080,0.4652}, {0.0147,0.2586,0.3533},
        {0.0049,0.3230,0.2720}, {0.0024,0.4073,0.2123}, {0.0093,0.5030,0.1582},
        {0.0291,0.6082,0.1117}, {0.0633,0.7100,0.0782}, {0.1096,0.7932,0.0573},
        {0.1655,0.8620,0.0422}, {0.2257,0.9149,0.0298}, {0.2904,0.9540,0.0203},
        {0.3597,0.9803,0.0134}, {0.4334,0.9950,0.0087}, {0.5121,1.0000,0.0057},
        {0.5945,0.9950,0.0039}, {0.6784,0.9786,0.0027}, {0.7621,0.9520,0.0021},
        {0.8425,0.9154,0.0018}, {0.9163,0.8700,0.0017}, {0.9786,0.8163,0.0014},
        {1.0263,0.7570,0.0011}, {1.0567,0.6949,0.0010}, {1.0622,0.6310,0.0008},
        {1.0456,0.5668,0.0006}, {1.0026,0.5030,0.0003}, {0.9384,0.4412,0.0002},
        {0.8544,0.3810,0.0002}, {0.7514,0.3210,0.0001}, {0.6424,0.2650,0.0000},
        {0.5419,0.2170,0.0000}, {0.4479,0.1750,0.0000}, {0.3608,0.1382,0.0000},
        {0.2835,0.1070,0.0000}, {0.2187,0.0816,0.0000}, {0.1649,0.0610,0.0000},
        {0.1212,0.0446,0.0000}, {0.0874,0.0320,0.0000}, {0.0636,0.0232,0.0000},
        {0.0468,0.0170,0.0000}, {0.0329,0.0119,0.0000}, {0.0227,0.0082,0.0000},
        {0.0158,0.0057,0.0000}, {0.0114,0.0041,0.0000}, {0.0081,0.0029,0.0000},
        {0.0058,0.0021,0.0000}, {0.0041,0.0015,0.0000}, {0.0029,0.0010,0.0000},
        {0.0020,0.0007,0.0000}, {0.0014,0.0005,0.0000}, {0.0010,0.0004,0.0000},
        {0.0007,0.0002,0.0000}, {0.0005,0.0002,0.0000}, {0.0003,0.0001,0.0000},
        {0.0002,0.0001,0.0000}, {0.0002,0.0001,0.0000}, {0.0001,0.0000,0.0000},
        {0.0001,0.0000,0.0000}, {0.0001,0.0000,0.0000}, {0.0000,0.0000,0.0000}
    };

    for (i = 0, lambda = 380; lambda < 780.1; i+=2, lambda += 10) {

        X += spectrum41.samples[i / 2] * cie_colour_match[i][0];
        Y += spectrum41.samples[i / 2] * cie_colour_match[i][1];
        Z += spectrum41.samples[i / 2] * cie_colour_match[i][2];

    }

        *x = X;
        *y = Y;
        *z = Z;
}

void DistantLight::xyz_to_rgb(double xc, double yc, double zc,
                double *r, double *g, double *b) const {
    double xr, yr, zr, xg, yg, zg, xb, yb, zb;
    double xw, yw, zw;
    double rx, ry, rz, gx, gy, gz, bx, by, bz;
    double rw, gw, bw;

	// CIE color system definitions
	double xRed = 0.7355;
	double yRed = 0.2645;
	double xGreen = 0.2658;
	double yGreen = 0.7243;
	double xBlue = 0.1669;
	double yBlue = 0.0085;
	double xWhite = 0.33333333;
	double yWhite = 0.33333333;
	double gamma = 0.0;

    xr = xRed;    yr = yRed;    zr = 1 - (xr + yr);
    xg = xGreen;  yg = yGreen;  zg = 1 - (xg + yg);
    xb = xBlue;   yb = yBlue;   zb = 1 - (xb + yb);

    xw = xWhite;  yw = yWhite;  zw = 1 - (xw + yw);

    /* xyz -> rgb matrix, before scaling to white. */
    
    rx = (yg * zb) - (yb * zg);  ry = (xb * zg) - (xg * zb);  rz = (xg * yb) - (xb * yg);
    gx = (yb * zr) - (yr * zb);  gy = (xr * zb) - (xb * zr);  gz = (xb * yr) - (xr * yb);
    bx = (yr * zg) - (yg * zr);  by = (xg * zr) - (xr * zg);  bz = (xr * yg) - (xg * yr);

    /* White scaling factors.
       Dividing by yw scales the white luminance to unity, as conventional. */
       
    if (fabs(yw) > 1.0e-48) {
      rw = ((rx * xw) + (ry * yw) + (rz * zw)) / yw;
      gw = ((gx * xw) + (gy * yw) + (gz * zw)) / yw;
      bw = ((bx * xw) + (by * yw) + (bz * zw)) / yw;
    } else {
      rw = 1.0;
      gw = 1.0;
      bw = 1.0;
    }

    /* xyz -> rgb matrix, correctly scaled to white. */
    
    rx = rx / rw;  ry = ry / rw;  rz = rz / rw;
    gx = gx / gw;  gy = gy / gw;  gz = gz / gw;
    bx = bx / bw;  by = by / bw;  bz = bz / bw;

    /* rgb of the desired point */
    
    *r = (rx * xc) + (ry * yc) + (rz * zc);
    *g = (gx * xc) + (gy * yc) + (gz * zc);
    *b = (bx * xc) + (by * yc) + (bz * zc);
}

void DistantLight::RGBtoHSV( float r, float g, float b, float *h, float *s, float *v ) const {
	float min, max, delta;

	if ( r < g ) {
		if ( r < b ) min = r; else min = b;
	} else {
		if ( g < b ) min = g; else min = b;
	}

	if ( r > g ) {
		if ( r > b ) max = r; else max = b;
	} else {
		if ( g > b ) max = g; else max = b;
	}

	*v = max;				// v

	delta = max - min;

	if( max != 0.f )
		*s = delta / max;		// s
	else {
		// r = g = b = 0		// s = 0, v is undefined
		*s = 0.f;
		*h = -1.f;
		return;
	}

	if( r == max )
		*h = ( g - b ) / delta;		// between yellow & magenta
	else if( g == max )
		*h = 2.f + ( b - r ) / delta;	// between cyan & yellow
	else
		*h = 4.f + ( r - g ) / delta;	// between magenta & cyan

	*h *= 60.f;				// degrees
	if( *h < 0 )
		*h += 360.f;

}

void DistantLight::HSVtoRGB( float *r, float *g, float *b, float h, float s, float v ) const {
	int i;
	float f, p, q, t;

	if( s == 0.f ) {
		// achromatic (grey)
		*r = *g = *b = v;
		return;
	}

	h /= 60.f;			// sector 0 to 5
	i = floor( h );
	f = h - i;			// factorial part of h
	p = v * ( 1.f - s );
	q = v * ( 1.f - s * f );
	t = v * ( 1.f - s * ( 1 - f ) );

	switch( i ) {
		case 0:
			*r = v;
			*g = t;
			*b = p;
			break;
		case 1:
			*r = q;
			*g = v;
			*b = p;
			break;
		case 2:
			*r = p;
			*g = v;
			*b = t;
			break;
		case 3:
			*r = p;
			*g = q;
			*b = v;
			break;
		case 4:
			*r = t;
			*g = p;
			*b = v;
			break;
		default:		// case 5:
			*r = v;
			*g = p;
			*b = q;
			break;
	}

}

extern "C" DLLEXPORT Light *CreateLight(const Transform &light2world,
		const ParamSet &paramSet) {
	Spectrum L = paramSet.FindOneSpectrum("L", Spectrum(1.0));
	float timeOfDay = paramSet.FindOneFloat("timeOfDay", 12.0f);
	float julianDay = paramSet.FindOneFloat("julianDay", 78.0f);
	float standardMeridian = paramSet.FindOneFloat("standardMeridian", 9.0f);
	float longitude = paramSet.FindOneFloat("longitude", 139.44f);
	float latitude = paramSet.FindOneFloat("latitude", 35.53f);
	float turb = paramSet.FindOneFloat("turbidity", 4.0f); 
	
	// Compute sun position
	
	float phiS, thetaS;
	float theta2, theta3, T, T2, chi;
	
	standardMeridian = radians(standardMeridian * 15);
	float solarTime = timeOfDay +
	(0.170*sin(4*M_PI*(julianDay - 80)/373) - 0.129*sin(2*M_PI*(julianDay - 8)/355)) +
	12.0 * (standardMeridian - radians(longitude))/M_PI;

    float solarDeclination = (0.4093*sin(2*M_PI*(julianDay - 81)/368));

    float solarAltitude= asin(sin(radians(latitude)) * sin(solarDeclination) -
			       cos(radians(latitude)) * cos(solarDeclination) * cos(M_PI*solarTime/12));

    float opp, adj;
    opp = -cos(solarDeclination) * sin(M_PI*solarTime/12);
    adj = -(cos(radians(latitude)) * sin(solarDeclination) +
	    sin(radians(latitude)) * cos(solarDeclination) *  cos(M_PI*solarTime/12));
    float solarAzimuth=atan2(opp,adj);

    phiS = -solarAzimuth;
    thetaS = M_PI / 2.0 - solarAltitude;

	Vector dir;
	dir.x = cos(phiS) * sin(thetaS);
	dir.y = sin(phiS) * sin(thetaS);
	dir.z = cos(thetaS);
		
	return new DistantLight(light2world, L, dir, turb, thetaS);
}
